@@ -27,26 +27,35 @@
 
 from __future__ import annotations
 
+import json
+import struct
+from dataclasses import dataclass, field
+from enum import Enum, EnumMeta, IntFlag, unique
+from typing import Any, Mapping, Sequence, cast
+
 from . import cbor
-from .cose import CoseKey, ES256
+from .cose import ES256, CoseKey
 from .utils import (
+    ByteBuffer,
+    _JsonDataObject,
     sha256,
     websafe_decode,
     websafe_encode,
-    ByteBuffer,
-    _CamelCaseDataObject,
 )
-from .features import webauthn_json_mapping
-from enum import Enum, EnumMeta, unique, IntFlag
-from dataclasses import dataclass, field
-from typing import Any, Mapping, Optional, Sequence, Tuple, Union, cast
-import struct
-import json
 
 """
 Data classes based on the W3C WebAuthn specification (https://www.w3.org/TR/webauthn/).
 
 See the specification for a description and details on their usage.
+
+Most of these classes can be serialized to JSON-compatible dictionaries by passing them
+to dict(), and then deserialized by calling DataClass.from_dict(data). For example:
+
+    user = PublicKeyCredentialUserEntity(id=b"1234", name="Alice")
+    data = dict(user)
+    # data is now a JSON-compatible dictionary, json.dumps(data) will work
+    user2 = PublicKeyCredentialUserEntity.from_dict(data)
+    assert user == user2
 """
 
 # Binary types
@@ -98,7 +107,7 @@ class AttestedCredentialData(bytes):
         return repr(self)
 
     @staticmethod
-    def _parse(data: bytes) -> Tuple[bytes, bytes, CoseKey, bytes]:
+    def _parse(data: bytes) -> tuple[bytes, bytes, CoseKey, bytes]:
         """Parse the components of an AttestedCredentialData from a binary
         string, and return them.
 
@@ -130,7 +139,7 @@ class AttestedCredentialData(bytes):
         )
 
     @classmethod
-    def unpack_from(cls, data: bytes) -> Tuple[AttestedCredentialData, bytes]:
+    def unpack_from(cls, data: bytes) -> tuple[AttestedCredentialData, bytes]:
         """Unpack an AttestedCredentialData from a byte string, returning it and
         any remaining data.
 
@@ -193,8 +202,8 @@ class AuthenticatorData(bytes):
     rp_id_hash: bytes
     flags: AuthenticatorData.FLAG
     counter: int
-    credential_data: Optional[AttestedCredentialData]
-    extensions: Optional[Mapping]
+    credential_data: AttestedCredentialData | None
+    extensions: Mapping | None
 
     def __init__(self, _: bytes):
         super().__init__()
@@ -230,7 +239,7 @@ class AuthenticatorData(bytes):
         flags: AuthenticatorData.FLAG,
         counter: int,
         credential_data: bytes = b"",
-        extensions: Optional[Mapping] = None,
+        extensions: Mapping | None = None,
     ):
         """Create an AuthenticatorData instance.
 
@@ -339,6 +348,7 @@ class CollectedClientData(bytes):
         CREATE = "webauthn.create"
         GET = "webauthn.get"
 
+    _data: Mapping[str, Any]
     type: str
     challenge: bytes
     origin: str
@@ -347,17 +357,17 @@ class CollectedClientData(bytes):
     def __init__(self, _: bytes):
         super().__init__()
 
-        data = json.loads(self.decode())
-        object.__setattr__(self, "type", data["type"])
-        object.__setattr__(self, "challenge", websafe_decode(data["challenge"]))
-        object.__setattr__(self, "origin", data["origin"])
-        object.__setattr__(self, "cross_origin", data.get("crossOrigin", False))
+        object.__setattr__(self, "_data", json.loads(self.decode()))
+        object.__setattr__(self, "type", self._data["type"])
+        object.__setattr__(self, "challenge", websafe_decode(self._data["challenge"]))
+        object.__setattr__(self, "origin", self._data["origin"])
+        object.__setattr__(self, "cross_origin", self._data.get("crossOrigin", False))
 
     @classmethod
     def create(
         cls,
         type: str,
-        challenge: Union[bytes, str],
+        challenge: bytes | str,
         origin: str,
         cross_origin: bool = False,
         **kwargs,
@@ -409,12 +419,6 @@ class _StringEnum(str, Enum, metaclass=_StringEnumMeta):
     """
 
 
-_b64_metadata = dict(
-    serialize=lambda x: websafe_encode(x) if webauthn_json_mapping.enabled else x,
-    deserialize=lambda x: websafe_decode(x) if webauthn_json_mapping.enabled else x,
-)
-
-
 @unique
 class AttestationConveyancePreference(_StringEnum):
     NONE = "none"
@@ -457,57 +461,54 @@ class PublicKeyCredentialType(_StringEnum):
     PUBLIC_KEY = "public-key"
 
 
-@dataclass(eq=False, frozen=True)
-class PublicKeyCredentialRpEntity(_CamelCaseDataObject):
+@unique
+class PublicKeyCredentialHint(_StringEnum):
+    SECURITY_KEY = "security-key"
+    CLIENT_DEVICE = "client-device"
+    HYBRID = "hybrid"
+
+
+def _as_cbor(data: _JsonDataObject) -> Mapping[str, Any]:
+    return {k: super(_JsonDataObject, data).__getitem__(k) for k in data}
+
+
+@dataclass(eq=False, frozen=True, kw_only=True)
+class PublicKeyCredentialRpEntity(_JsonDataObject):
     name: str
-    id: Optional[str] = None
+    id: str | None = None
 
     @property
-    def id_hash(self) -> Optional[bytes]:
+    def id_hash(self) -> bytes | None:
         """Return SHA256 hash of the identifier."""
         return sha256(self.id.encode("utf8")) if self.id else None
 
 
-@dataclass(eq=False, frozen=True)
-class PublicKeyCredentialUserEntity(_CamelCaseDataObject):
+@dataclass(eq=False, frozen=True, kw_only=True)
+class PublicKeyCredentialUserEntity(_JsonDataObject):
     name: str
-    id: bytes = field(metadata=_b64_metadata)
-    display_name: Optional[str] = None
+    id: bytes
+    display_name: str | None = None
 
 
-@dataclass(eq=False, frozen=True)
-class PublicKeyCredentialParameters(_CamelCaseDataObject):
+@dataclass(eq=False, frozen=True, kw_only=True)
+class PublicKeyCredentialParameters(_JsonDataObject):
     type: PublicKeyCredentialType
     alg: int
 
-    @classmethod
-    def _deserialize_list(cls, value):
-        if value is None:
-            return None
-        items = [cls.from_dict(e) for e in value]
-        return [e for e in items if e.type is not None]
 
-
-@dataclass(eq=False, frozen=True)
-class PublicKeyCredentialDescriptor(_CamelCaseDataObject):
+@dataclass(eq=False, frozen=True, kw_only=True)
+class PublicKeyCredentialDescriptor(_JsonDataObject):
     type: PublicKeyCredentialType
-    id: bytes = field(metadata=_b64_metadata)
-    transports: Optional[Sequence[AuthenticatorTransport]] = None
-
-    @classmethod
-    def _deserialize_list(cls, value):
-        if value is None:
-            return None
-        items = [cls.from_dict(e) for e in value]
-        return [e for e in items if e.type is not None]
+    id: bytes
+    transports: Sequence[AuthenticatorTransport] | None = None
 
 
-@dataclass(eq=False, frozen=True)
-class AuthenticatorSelectionCriteria(_CamelCaseDataObject):
-    authenticator_attachment: Optional[AuthenticatorAttachment] = None
-    resident_key: Optional[ResidentKeyRequirement] = None
-    user_verification: Optional[UserVerificationRequirement] = None
-    require_resident_key: Optional[bool] = False
+@dataclass(eq=False, frozen=True, kw_only=True)
+class AuthenticatorSelectionCriteria(_JsonDataObject):
+    authenticator_attachment: AuthenticatorAttachment | None = None
+    resident_key: ResidentKeyRequirement | None = None
+    user_verification: UserVerificationRequirement | None = None
+    require_resident_key: bool | None = False
 
     def __post_init__(self):
         super().__post_init__()
@@ -529,121 +530,168 @@ class AuthenticatorSelectionCriteria(_CamelCaseDataObject):
         )
 
 
-@dataclass(eq=False, frozen=True)
-class PublicKeyCredentialCreationOptions(_CamelCaseDataObject):
+@dataclass(eq=False, frozen=True, kw_only=True)
+class PublicKeyCredentialCreationOptions(_JsonDataObject):
     rp: PublicKeyCredentialRpEntity
     user: PublicKeyCredentialUserEntity
-    challenge: bytes = field(metadata=_b64_metadata)
-    pub_key_cred_params: Sequence[PublicKeyCredentialParameters] = field(
-        metadata=dict(deserialize=PublicKeyCredentialParameters._deserialize_list),
-    )
-    timeout: Optional[int] = None
-    exclude_credentials: Optional[Sequence[PublicKeyCredentialDescriptor]] = field(
-        default=None,
-        metadata=dict(deserialize=PublicKeyCredentialDescriptor._deserialize_list),
-    )
-    authenticator_selection: Optional[AuthenticatorSelectionCriteria] = None
-    attestation: Optional[AttestationConveyancePreference] = None
-    extensions: Optional[Mapping[str, Any]] = None
+    challenge: bytes
+    pub_key_cred_params: Sequence[PublicKeyCredentialParameters]
+    timeout: int | None = None
+    exclude_credentials: Sequence[PublicKeyCredentialDescriptor] | None = None
+    authenticator_selection: AuthenticatorSelectionCriteria | None = None
+    hints: Sequence[PublicKeyCredentialHint] | None = None
+    attestation: AttestationConveyancePreference | None = None
+    attestation_formats: Sequence[str] | None = None
+    extensions: Mapping[str, Any] | None = None
 
 
-@dataclass(eq=False, frozen=True)
-class PublicKeyCredentialRequestOptions(_CamelCaseDataObject):
-    challenge: bytes = field(metadata=_b64_metadata)
-    timeout: Optional[int] = None
-    rp_id: Optional[str] = None
-    allow_credentials: Optional[Sequence[PublicKeyCredentialDescriptor]] = field(
-        default=None,
-        metadata=dict(deserialize=PublicKeyCredentialDescriptor._deserialize_list),
-    )
-    user_verification: Optional[UserVerificationRequirement] = None
-    extensions: Optional[Mapping[str, Any]] = None
+@dataclass(eq=False, frozen=True, kw_only=True)
+class PublicKeyCredentialRequestOptions(_JsonDataObject):
+    challenge: bytes
+    timeout: int | None = None
+    rp_id: str | None = None
+    allow_credentials: Sequence[PublicKeyCredentialDescriptor] | None = None
+    user_verification: UserVerificationRequirement | None = None
+    hints: Sequence[PublicKeyCredentialHint] | None = None
+    extensions: Mapping[str, Any] | None = None
 
 
-@dataclass(eq=False, frozen=True)
-class AuthenticatorAttestationResponse(_CamelCaseDataObject):
-    client_data: CollectedClientData = field(
-        metadata=dict(
-            _b64_metadata,
-            name="clientDataJSON",
-        )
-    )
-    attestation_object: AttestationObject = field(metadata=_b64_metadata)
-    extension_results: Optional[Mapping[str, Any]] = None
+@dataclass(eq=False, frozen=True, kw_only=True)
+class AuthenticatorAttestationResponse(_JsonDataObject):
+    client_data: CollectedClientData = field(metadata=dict(name="clientDataJSON"))
+    attestation_object: AttestationObject
+
+
+@dataclass(eq=False, frozen=True, kw_only=True)
+class AuthenticatorAssertionResponse(_JsonDataObject):
+    client_data: CollectedClientData = field(metadata=dict(name="clientDataJSON"))
+    authenticator_data: AuthenticatorData
+    signature: bytes
+    user_handle: bytes | None = None
+
+
+class AuthenticationExtensionsClientOutputs(Mapping[str, Any]):
+    """Holds extension output from a call to MakeCredential or GetAssertion.
+
+    When accessed as a dict, all bytes values will be serialized to base64url encoding,
+    capable of being serialized to JSON.
+
+    When accessed using attributes, richer types will instead be returned.
+    """
+
+    def __init__(self, outputs: Mapping[str, Any] = {}):
+        self._members = {k: v for k, v in outputs.items() if v is not None}
+
+    def __iter__(self):
+        return iter(self._members)
+
+    def __len__(self):
+        return len(self._members)
 
     def __getitem__(self, key):
-        if key == "clientData" and not webauthn_json_mapping.enabled:
-            return self.client_data
-        return super().__getitem__(key)
+        value = self._members[key]
+        if isinstance(value, bytes):
+            return websafe_encode(value)
+        elif isinstance(value, Mapping) and not isinstance(value, dict):
+            return dict(value)
+        return value
 
-    @classmethod
-    def from_dict(cls, data: Optional[Mapping[str, Any]]):
-        if data is not None and not webauthn_json_mapping.enabled:
-            value = dict(data)
-            value["clientDataJSON"] = value.pop("clientData", None)
-            data = value
-        return super().from_dict(data)
+    def __getattr__(self, key):
+        parts = key.split("_")
+        name = parts[0] + "".join(p.title() for p in parts[1:])
+        return self._members.get(name)
+
+    def __repr__(self):
+        return repr(dict(self))
 
 
-@dataclass(eq=False, frozen=True)
-class AuthenticatorAssertionResponse(_CamelCaseDataObject):
-    client_data: CollectedClientData = field(
-        metadata=dict(
-            _b64_metadata,
-            name="clientDataJSON",
-        )
-    )
-    authenticator_data: AuthenticatorData = field(metadata=_b64_metadata)
-    signature: bytes = field(metadata=_b64_metadata)
-    user_handle: Optional[bytes] = field(metadata=_b64_metadata, default=None)
-    credential_id: Optional[bytes] = field(metadata=_b64_metadata, default=None)
-    extension_results: Optional[Mapping[str, Any]] = None
+@dataclass(eq=False, frozen=True, kw_only=True)
+class RegistrationResponse(_JsonDataObject):
+    """
+    Represents the RegistrationResponse structure from the WebAuthn specification,
+    with fields modeled after the JSON serialization.
 
-    def __getitem__(self, key):
-        if key == "clientData" and not webauthn_json_mapping.enabled:
-            return self.client_data
-        return super().__getitem__(key)
+    Serializing this object to JSON can be done by using json.dumps(dict(response)).
 
-    @classmethod
-    def from_dict(cls, data: Optional[Mapping[str, Any]]):
-        if data is not None and not webauthn_json_mapping.enabled:
-            value = dict(data)
-            value["clientDataJSON"] = value.pop("clientData", None)
-            data = value
-        return super().from_dict(data)
+    See: https://www.w3.org/TR/webauthn-3/#dictdef-registrationresponsejson
+    """
 
-
-@dataclass(eq=False, frozen=True)
-class RegistrationResponse(_CamelCaseDataObject):
-    id: bytes = field(metadata=_b64_metadata)
+    id: str = field(init=False)
+    raw_id: bytes
     response: AuthenticatorAttestationResponse
-    authenticator_attachment: Optional[AuthenticatorAttachment] = None
-    client_extension_results: Optional[Mapping] = None
-    type: Optional[PublicKeyCredentialType] = None
+    authenticator_attachment: AuthenticatorAttachment | None = None
+    client_extension_results: AuthenticationExtensionsClientOutputs = field(
+        default_factory=AuthenticationExtensionsClientOutputs
+    )
+    type: PublicKeyCredentialType = PublicKeyCredentialType.PUBLIC_KEY
 
     def __post_init__(self):
-        webauthn_json_mapping.require()
+        object.__setattr__(self, "id", websafe_encode(self.raw_id))
         super().__post_init__()
 
+    @classmethod
+    def _parse_value(cls, t, value):
+        if t == Mapping[str, Any] | None:
+            # Don't convert extension_results
+            return value
+        return super()._parse_value(t, value)
 
-@dataclass(eq=False, frozen=True)
-class AuthenticationResponse(_CamelCaseDataObject):
-    id: bytes = field(metadata=_b64_metadata)
+    @classmethod
+    def from_dict(cls, data):
+        if data and "id" in data:
+            data = dict(data)
+            credential_id = data.pop("id")
+            if credential_id != data["rawId"]:
+                raise ValueError("id does not match rawId")
+        return super().from_dict(data)
+
+
+@dataclass(eq=False, frozen=True, kw_only=True)
+class AuthenticationResponse(_JsonDataObject):
+    """
+    Represents the AuthenticationResponse structure from the WebAuthn specification,
+    with fields modeled after the JSON serialization.
+
+    Serializing this object to JSON can be done by using json.dumps(dict(response)).
+
+    See: https://www.w3.org/TR/webauthn-3/#dictdef-authenticationresponsejson
+    """
+
+    id: str = field(init=False)
+    raw_id: bytes
     response: AuthenticatorAssertionResponse
-    authenticator_attachment: Optional[AuthenticatorAttachment] = None
-    client_extension_results: Optional[Mapping] = None
-    type: Optional[PublicKeyCredentialType] = None
+    authenticator_attachment: AuthenticatorAttachment | None = None
+    client_extension_results: AuthenticationExtensionsClientOutputs = field(
+        default_factory=AuthenticationExtensionsClientOutputs
+    )
+    type: PublicKeyCredentialType = PublicKeyCredentialType.PUBLIC_KEY
 
     def __post_init__(self):
-        webauthn_json_mapping.require()
+        object.__setattr__(self, "id", websafe_encode(self.raw_id))
         super().__post_init__()
 
+    @classmethod
+    def _parse_value(cls, t, value):
+        if t == Mapping[str, Any] | None:
+            # Don't convert extension_results
+            return value
+        return super()._parse_value(t, value)
 
-@dataclass(eq=False, frozen=True)
-class CredentialCreationOptions(_CamelCaseDataObject):
+    @classmethod
+    def from_dict(cls, data):
+        if data and "id" in data:
+            data = dict(data)
+            credential_id = data.pop("id")
+            if credential_id != data["rawId"]:
+                raise ValueError("id does not match rawId")
+        return super().from_dict(data)
+
+
+@dataclass(eq=False, frozen=True, kw_only=True)
+class CredentialCreationOptions(_JsonDataObject):
     public_key: PublicKeyCredentialCreationOptions
 
 
-@dataclass(eq=False, frozen=True)
-class CredentialRequestOptions(_CamelCaseDataObject):
+@dataclass(eq=False, frozen=True, kw_only=True)
+class CredentialRequestOptions(_JsonDataObject):
     public_key: PublicKeyCredentialRequestOptions
